from django.conf.urls import patterns, url
from django.views.generic import TemplateView

urlpatterns = patterns('',
    url(r'^preview.html$', TemplateView.as_view(template_name='faker/preview.html'))
)


from faker import Faker
from django_faker.populator import Populator
from django_faker import Faker as DjangoFaker

from django.db import models
from django.utils import unittest
from django.template import Context, TemplateSyntaxError
from django.template import Template

fake = Faker()

class Game(models.Model):

    title= models.CharField(max_length=200)
    slug= models.SlugField(max_length=200)
    description= models.TextField()
    created_at= models.DateTimeField()
    updated_date= models.DateField()
    updated_time= models.TimeField()
    active= models.BooleanField()
    max_score= models.BigIntegerField()


class Player(models.Model):

    nickname= models.CharField(max_length=100)
    score= models.BigIntegerField()
    last_login_at= models.DateTimeField()

    game= models.ForeignKey(Game)


class Action(models.Model):

    ACTION_FIRE='fire'
    ACTION_MOVE='move'
    ACTION_STOP='stop'


    ACTIONS = (
        (ACTION_FIRE,'Fire'),
        (ACTION_MOVE,'Move'),
        (ACTION_STOP,'Stop'),
        )

    name= models.CharField(max_length=4, choices=ACTIONS)
    executed_at= models.DateTimeField()

    actor= models.ForeignKey(Player,related_name='actions', null=True)
    target= models.ForeignKey(Player, related_name='enemy_actions+', null=True)


class PopulatorTestCase(unittest.TestCase):

    def testPopulation(self):

        generator = fake
        populator = Populator(generator)
        populator.addEntity( Game, 10 )
        self.assertEqual(len(populator.execute()[Game]), 10)

    def testGuesser(self):

        generator = fake

        def title_fake(arg):
            title_fake.count += 1
            name = generator.company()
            return name
        title_fake.count = 0

        populator = Populator(generator)
        populator.addEntity( Game, 10, {
            'title': title_fake
        } )
        self.assertEqual(len(populator.execute()[Game]), title_fake.count)

    def testFormatter(self):

        generator = fake

        populator = Populator( generator )

        populator.addEntity(Game,5)
        populator.addEntity(Player, 10, {
            'score': lambda x: fake.randomInt(0,1000),
            'nickname': lambda x: fake.email()
        })
        populator.addEntity(Action,30)

        insertedPks = populator.execute()

        self.assertTrue( len(insertedPks[Game]) == 5 )
        self.assertTrue( len(insertedPks[Player]) == 10 )

        self.assertTrue( any([0 <= p.score <= 1000 and '@' in p.nickname for p in Player.objects.all() ]) )


class TemplateTagsTestCase(unittest.TestCase):

    @staticmethod
    def render( template, context=None ):
        t = Template("{% load fakers %}"+template)
        c = Context(context or {})
        text=  t.render(c)
        return text

    # do_fake: fake
    def testSimpleFakeTag(self):
        self.assertNotEqual(self.render("{% fake 'name' as myname %}{{ myname }}"),"")

    def testSimpleFakeTagWithArguments(self):
        self.assertNotEqual(self.render("{% fake 'dateTimeBetween' '-10d' as mydate %}{{ mydate }}"),"")

    def testSimpleFakeTagFormatterNotFoundRaisesException(self):
        with self.assertRaises(AttributeError):
            self.render("{% fake 'notFoundedFake' as foo %}")

    def testSimpleFakeTagOptionalAssignment(self):
        self.assertNotEqual(self.render("{% fake 'name' %}"),"")
        self.assertEqual(len(self.render("{% fake 'md5' %}")),32)

    # do_fake_filter: fake
    def testFakeFilterTag(self):
        self.assertIn(self.render("{{ 'randomElement'|fake:'testString' }}"),'testString')

    def testFakeFilterWithValueFromContext(self):
        mylist = [100,200,300]
        rendered = self.render("{{ 'randomElement'|fake:mylist }}", {'mylist': mylist})
        self.assertIn(rendered, [unicode(el) for el in mylist])

    def testFakeFilterFormatterNotFoundRaisesException(self):
        with self.assertRaises(AttributeError):
            self.render("{{ 'notFoundedFake'|fake:mylist }}", {'mylist': [100,200,300]})

    def testFakeFilterAsIfCondition(self):
        self.assertEqual(self.render("{% if 'boolean'|fake:100 %}True forever{% endif %}"), "True forever")
        self.assertEqual(self.render("{% if 'boolean'|fake:0 %}{% else %}False forever{% endif %}"), "False forever")

    def testFakeFilterAsForInRange(self):
        times = 10
        rendered = self.render("{% for word in 'words'|fake:times %}{{ word }}\n{% endfor %}",{'times':times})
        words = [word for word in rendered.split('\n') if word.strip()]
        self.assertEqual( len(words) , times)

    # do_or_fake_filter: or_fake
    def testOrFakeFilterTag(self):
        self.assertEqual(len(self.render("{{ unknown_var|or_fake:'md5' }}")), 32)


    def testFullXmlContact(self):
        self.assertTrue(self.render("""<?xml version="1.0" encoding="UTF-8"?>
<contacts>
    {% fake 'randomInt' 10 20 as times %}
    {% for i in 10|get_range %}
    <contact firstName="{% fake 'firstName' %}" lastName="{% fake 'lastName' %}" email="{% fake 'email' %}"/>
        <phone number="{% fake 'phoneNumber' %}"/>
        {% if 'boolean'|fake:25 %}
        <birth date="{{ 'dateTimeThisCentury'|fake|date:"D d M Y" }}" place="{% fake 'city' %}"/>
        {% endif %}
        <address>
            <street>{% fake 'streetAddress' %}</street>
            <city>{% fake 'city' %}</city>
            <postcode>{% fake 'postcode' %}</postcode>
            <state>{% fake 'state' %}</state>
        </address>
        <company name="{% fake 'company' %}" catchPhrase="{% fake 'catchPhrase' %}">
        {% if 'boolean'|fake:25 %}
            <offer>{% fake 'bs' %}</offer>
        {% endif %}
        {% if 'boolean'|fake:33 %}
            <director name="{% fake 'name' %}" />
        {% endif %}
        </company>
        {% if 'boolean'|fake:15 %}
        <details>
        <![CDATA[
        {% fake 'text' 500 %}
        ]]>
        </details>
        {% endif %}
    </contact>
    {% endfor %}
</contacts>
"""))


class APIDjangoFakerTestCase(unittest.TestCase):

    def testDjangoFakerSingleton(self):

        self.assertEqual( DjangoFaker() , DjangoFaker() )
        self.assertIs( DjangoFaker() , DjangoFaker() )

    def testFakerCacheGenerator(self):

        self.assertEqual( DjangoFaker().getGenerator(), DjangoFaker().getGenerator() )
        self.assertIs( DjangoFaker().getGenerator(), DjangoFaker().getGenerator() )
        self.assertIs( DjangoFaker().getGenerator(codename='default'), DjangoFaker().getGenerator(codename='default') )

        self.assertEqual( DjangoFaker().getGenerator(locale='it_IT'), DjangoFaker().getGenerator(locale='it_IT') )
        self.assertIs( DjangoFaker().getGenerator(locale='it_IT'), DjangoFaker().getGenerator(locale='it_IT') )

    def testFakerCachePopulator(self):

        self.assertEqual( DjangoFaker().getPopulator(), DjangoFaker().getPopulator() )
        self.assertIs( DjangoFaker().getPopulator(), DjangoFaker().getPopulator() )
        self.assertIs( DjangoFaker().getPopulator().generator, DjangoFaker().getPopulator().generator )

        self.assertEqual( DjangoFaker().getPopulator(locale='it_IT'), DjangoFaker().getPopulator(locale='it_IT') )
        self.assertIs( DjangoFaker().getPopulator(locale='it_IT'), DjangoFaker().getPopulator(locale='it_IT') )
        


"""

Django-faker uses python-faker to generate test data for Django models and templates.

"""
from django.conf import settings

__version__ = '0.2'

class Faker(object):

    instance = None
    populators = {}
    generators = {}

    @classmethod
    def __new__(cls, *args, **kwargs):
        if cls.instance is None:
            cls.instance = super(Faker, cls).__new__(*args, **kwargs)
        return cls.instance

    def __init__(self):
#        assert False, "Cannot create a instance of Faker"
        pass


    @staticmethod
    def getCodename(locale=None, providers=None):
        """
        codename = locale[-Provider]*
        """
        # language
        locale = locale or getattr(settings,'FAKER_LOCALE', getattr(settings,'LANGUAGE_CODE', None))
        # providers
        providers = providers or getattr(settings,'FAKER_PROVIDERS', None)

        codename = locale or 'default'

        if providers:
            codename += "-" + "-".join(sorted(providers))

        return codename


    @classmethod
    def getGenerator(cls, locale=None, providers=None, codename=None):
        """
        use a codename to cache generators
        """

        codename = codename or cls.getCodename(locale, providers)

        if codename not in cls.generators:
            from faker import Faker as FakerGenerator
            # initialize with faker.generator.Generator instance
            # and remember in cache
            cls.generators[codename] = FakerGenerator( locale, providers )
            cls.generators[codename].seed( cls.generators[codename].randomInt() )

        return cls.generators[codename]



    @classmethod
    def getPopulator(cls, locale=None, providers=None):
        """

        uses:

            from django_faker import Faker
            pop = Faker.getPopulator()

            from myapp import models
            pop.addEntity(models.MyModel, 10)
            pop.addEntity(models.MyOtherModel, 10)
            pop.execute()

            pop = Faker.getPopulator('it_IT')

            pop.addEntity(models.MyModel, 10)
            pop.addEntity(models.MyOtherModel, 10)
            pop.execute()

        """

        codename = cls.getCodename(locale, providers)

        if codename not in cls.populators:

            generator = cls.generators.get(codename,  None) or cls.getGenerator(codename=codename)

            from django_faker import populator

            cls.populators[codename] = populator.Populator( generator )

        return cls.populators[codename]

#        if not cls.populator:
#            cls.populator= populators.Populator(
#            # initialize with faker.generator.Generator instance
#            FakerGenerator(
#
#                getattr(settings,'FAKER_LOCALE', getattr(settings,'LANGUAGE_CODE', locale)),
#
#                getattr(settings,'FAKER_PROVIDERS', providers)
#            )
#        )




from inspect import getargspec
from django import template
from django.template.base import TagHelperNode, TemplateSyntaxError, parse_bits

register = template.Library()

from django_faker import Faker

def optional_assignment_tag(func=None, takes_context=None, name=None):
    """
    https://groups.google.com/forum/?fromgroups=#!topic/django-developers/E0XWFrkRMGc
    new template tags type
    """
    def dec(func):
        params, varargs, varkw, defaults = getargspec(func)

        class AssignmentNode(TagHelperNode):
            def __init__(self, takes_context, args, kwargs, target_var=None):
                super(AssignmentNode, self).__init__(takes_context, args, kwargs)
                self.target_var = target_var

            def render(self, context):
                resolved_args, resolved_kwargs = self.get_resolved_arguments(context)
                output = func(*resolved_args, **resolved_kwargs)
                if self.target_var is None:
                    return output
                else:
                    context[self.target_var] = output
                return ''

        function_name = (name or
                         getattr(func, '_decorated_function', func).__name__)

        def compile_func(parser, token):
            bits = token.split_contents()[1:]
            if len(bits) < 2 or bits[-2] != 'as':
                target_var = None
            else:
                target_var = bits[-1]
                bits = bits[:-2]
            args, kwargs = parse_bits(parser, bits, params,
                varargs, varkw, defaults, takes_context, function_name)
            return AssignmentNode(takes_context, args, kwargs, target_var)

        compile_func.__doc__ = func.__doc__
        register.tag(function_name, compile_func)
        return func
    if func is None:
        # @register.assignment_tag(...)
        return dec
    elif callable(func):
        # @register.assignment_tag
        return dec(func)
    else:
        raise TemplateSyntaxError("Invalid arguments provided to assignment_tag")

@optional_assignment_tag(name='fake')
def do_fake( formatter, *args, **kwargs ):
    """
        call a faker format
        uses:

            {% fake "formatterName" *args **kwargs as myvar %}
            {{ myvar }}

        or:
            {% fake 'name' %}

        """
    return Faker.getGenerator().format( formatter, *args, **kwargs )


#@register.assignment_tag(name='fake')
#def fake_tag_as( formatter, *args, **kwargs ):
#    """
#    call a faker format
#    uses:
#
#        {% fake "formatterName" *args **kwargs as myvar %}
#
#    """
#    return Faker.getGenerator().format( formatter, *args, **kwargs )

#@register.simple_tag(name='fakestr')
#def fake_tag_str( formatter, *args, **kwargs ):
#    """
#    call a faker format
#    uses:
#
#        {% fakestr "formatterName" *values **kwargs %}
#    """
#    if formatter == 'dateTimeThisCentury' : print args, kwargs
#    return Faker.getGenerator().format( formatter, *args, **kwargs )

@register.filter(name='fake')
def do_fake_filter( formatter, arg=None ):
    """
    call a faker format
    uses:

        {{ 'randomElement'|fake:mylist }}
        {% if 'boolean'|fake:30 %} .. {% endif %}
        {% for word in 'words'|fake:times %}{{ word }}\n{% endfor %}

    """
    args = []
    if not arg is None: args.append(arg)
    return Faker.getGenerator().format( formatter, *args )


@register.filter(name='or_fake')
def do_or_fake_filter( value, formatter ):
    """
    call a faker if value is None
    uses:

        {{ myint|or_fake:'randomInt' }}

    """
    if not value:
        value = Faker.getGenerator().format( formatter )
    return value


@register.filter
def get_range( value ):
    """
        http://djangosnippets.org/snippets/1357/

      Filter - returns a list containing range made from given value
      Usage (in template):

      <ul>{% for i in 3|get_range %}
        <li>{{ i }}. Do something</li>
      {% endfor %}</ul>

      Results with the HTML:
      <ul>
        <li>0. Do something</li>
        <li>1. Do something</li>
        <li>2. Do something</li>
      </ul>

      Instead of 3 one may use the variable set in the views
    """
    return range( value )


import re

class Name(object):
    def __init__(self, generator):
        """
        :param generator Generator
        """
        self.generator = generator

    def guessFormat(self, name):
        """
        :param name:
        :type name: str
        """
        name = name.lower()
        generator = self.generator
        if re.findall(r'^is[_A-Z]', name): return lambda x:generator.boolean()
        if re.findall(r'(_a|A)t$', name): return lambda x:generator.dateTime()

        if name in ('first_name','firstname'): return lambda x: generator.firstName()
        if name in ('last_name','lastname'): return lambda x: generator.lastName()

        if name in ('username','login','nickname'): return lambda x:generator.userName()
        if name in ('email','email_address'): return lambda x:generator.email()
        if name in ('phone_number','phonenumber','phone'): return lambda x:generator.phoneNumber()
        if name == 'address' : return lambda x:generator.address()
        if name == 'city' : return lambda x: generator.city()
        if name == 'streetaddress' : return lambda x: generator.streetaddress()
        if name in ('postcode','zipcode'): return lambda x: generator.postcode()
        if name == 'state' : return lambda x: generator.state()
        if name == 'country' : return lambda x: generator.country()
        if name == 'title' : return lambda x: generator.sentence()
        if name in ('body','summary', 'description'): return lambda x: generator.text()


import os
from setuptools import setup, find_packages


def read_file(filename):
    """Read a file into a string"""
    path = os.path.abspath(os.path.dirname(__file__))
    filepath = os.path.join(path, filename)
    try:
        return open(filepath).read()
    except IOError:
        return ''


setup(
    name='django-faker',
    version=__import__('django_faker').__version__,
    author='joke2k',
    author_email='joke2k@gmail.com',
    packages=find_packages(),
    include_package_data=True,
    url='http://github.com/joke2k/django-faker',
    license='MIT',
    description=u' '.join(__import__('django_faker').__doc__.splitlines()).strip(),
    classifiers=[
        # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Information Technology',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Testing',
        'Topic :: Software Development :: Widget Sets',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
        'Framework :: Django'
    ],
    keywords='faker fixtures data test django',
    long_description=read_file('README.rst'),
    install_requires=['django','fake-factory>=0.2'],
    tests_require=['django','fake-factory>=0.2'],
    test_suite="runtests.runtests",
    zip_safe=False,
)


import random
from django_faker.guessers import Name
from django.db.models.fields import *
from django.db.models import ForeignKey, ManyToManyField, OneToOneField, ImageField


class FieldTypeGuesser(object):

    def __init__(self, generator):
        """
        :param generator: Generator
        """
        self.generator = generator

    def guessFormat(self, field):

        generator = self.generator
        if isinstance(field, BooleanField): return lambda x: generator.boolean()
        if isinstance(field, NullBooleanField): return lambda x: generator.nullBoolean()
        if isinstance(field, DecimalField): return lambda x: generator.pydecimal(rightDigits=field.decimal_places)
        if isinstance(field, SmallIntegerField): return lambda x: generator.randomInt(0,65535)
        if isinstance(field, IntegerField): return lambda x: generator.randomInt(0,4294967295)
        if isinstance(field, BigIntegerField): return lambda x: generator.randomInt(0,18446744073709551615)
        if isinstance(field, FloatField): return lambda x: generator.pyfloat()
        if isinstance(field, CharField):
            if field.choices:
                return lambda x: generator.randomElement(field.choices)[0]
            return lambda x: generator.text(field.max_length) if field.max_length >= 5 else generator.word()
        if isinstance(field, TextField): return lambda x: generator.text()

        if isinstance(field, DateTimeField): return lambda x: generator.dateTime()
        if isinstance(field, DateField): return lambda x: generator.date()
        if isinstance(field, TimeField): return lambda x: generator.time()

        if isinstance(field, URLField): return lambda x: generator.uri()
        if isinstance(field, SlugField): return lambda x: generator.slug()
        if isinstance(field, IPAddressField):
            protocolIp = generator.randomElements(['ipv4','ipv6'])
            return lambda x: getattr(generator,protocolIp)()
        if isinstance(field, EmailField): return lambda x: generator.email()
        if isinstance(field, ImageField): return lambda x: None

        raise AttributeError(field)


class ModelPopulator(object):
    def __init__(self, model):
        """
        :param model: Generator
        """
        self.model = model
        self.fieldFormatters = {}

    def guessFieldFormatters(self, generator):

        formatters = {}
        model = self.model
        nameGuesser = Name(generator)
        fieldTypeGuesser = FieldTypeGuesser(generator)

        for field in model._meta.fields:
        #            yield field.name, getattr(self, field.name)
            fieldName = field.name
            if isinstance(field, (ForeignKey,ManyToManyField,OneToOneField)):
                relatedModel = field.rel.to

                def build_relation(inserted):
                    if relatedModel in inserted and inserted[relatedModel]:
                        return relatedModel.objects.get(pk=random.choice(inserted[relatedModel]))
                    if not field.null:
                        try :
                            # try to retrieve random object from relatedModel
                            relatedModel.objects.order_by('?')[0]
                        except IndexError:
                            raise Exception('Relation "%s.%s" with "%s" cannot be null, check order of addEntity list' % (
                                field.model.__name__, field.name, relatedModel.__name__,
                            ))
                    return None

                formatters[fieldName] = build_relation
                continue

            if isinstance(field, AutoField):
                continue

            formatter = nameGuesser.guessFormat(fieldName)
            if formatter:
                formatters[fieldName] = formatter
                continue

            formatter = fieldTypeGuesser.guessFormat(field)
            if formatter:
                formatters[fieldName] = formatter
                continue

        return formatters

    def execute(self, using, insertedEntities):

        obj = self.model()

        for field, format in self.fieldFormatters.items():
            if format:
                value = format(insertedEntities) if hasattr(format,'__call__') else format
                setattr(obj, field, value)

        obj.save(using=using)

        return obj.pk


class Populator(object):

    def __init__(self, generator):
        """
        :param generator: Generator
        """
        self.generator = generator
        self.entities = {}
        self.quantities = {}
        self.orders = []


    def addEntity(self, model, number, customFieldFormatters=None):
        """
        Add an order for the generation of $number records for $entity.

        :param model: mixed A Django Model classname, or a faker.orm.django.EntityPopulator instance
        :type model: Model
        :param number: int The number of entities to populate
        :type number: integer
        :param customFieldFormatters: optional dict with field as key and callable as value
        :type customFieldFormatters: dict or None
        """
        if not isinstance(model, ModelPopulator):
            model = ModelPopulator(model)

        model.fieldFormatters = model.guessFieldFormatters( self.generator )
        if customFieldFormatters:
            model.fieldFormatters.update(customFieldFormatters)

        klass = model.model
        self.entities[klass] = model
        self.quantities[klass] = number
        self.orders.append(klass)

    def execute(self, using=None):
        """
        Populate the database using all the Entity classes previously added.

        :param using A Django database connection name
        :rtype: A list of the inserted PKs
        """
        if not using:
            using = self.getConnection()

        insertedEntities = {}
        for klass in self.orders:
            number = self.quantities[klass]
            if klass not in insertedEntities:
                insertedEntities[klass] = []
            for i in range(0,number):
                    insertedEntities[klass].append( self.entities[klass].execute(using, insertedEntities) )

        return insertedEntities

    def getConnection(self):
        """
        use the first connection available
        :rtype: Connection
        """

        klass = self.entities.keys()
        if not klass:
            raise AttributeError('No class found from entities. Did you add entities to the Populator ?')
        klass = klass[0]

        return klass.objects._db



